@@ -42,8 +42,10 @@ struct _FrdpSessionPrivate
 
   GtkWidget    *display;
   cairo_surface_t *surface;
+  cairo_format_t cairo_format;
   gboolean scaling;
-  double scale;
+  double scale_x;
+  double scale_y;
   double offset_x;
   double offset_y;
 
@@ -76,6 +78,7 @@ enum
 
 enum
 {
+  RDP_ERROR,
   RDP_CONNECTED,
   RDP_DISCONNECTED,
   RDP_AUTH_FAILURE,
@@ -124,23 +127,22 @@ frdp_session_update_mouse_pointer (FrdpSession  *self)
     cursor = gdk_cursor_new_from_name (display, "default");
   else {
     rdpPointer *pointer = &priv->cursor->pointer;
-    double scale = self->priv->scale;
-    double x = priv->cursor->pointer.xPos * scale;
-    double y = priv->cursor->pointer.yPos * scale;
-    double w = pointer->width * scale;
-    double h = pointer->height * scale;
+    double x = priv->cursor->pointer.xPos * priv->scale_x;
+    double y = priv->cursor->pointer.yPos * priv->scale_y;
+    double w = pointer->width * priv->scale_x;
+    double h = pointer->height * priv->scale_y;
     cairo_surface_t *surface;
     cairo_t *cairo;
 
     if (!self->priv->scaling) {
-      scale = 1.0;
+      self->priv->scale_x = self->priv->scale_y = 1.0;
     }
 
     /* Scale the source image according to current settings. */
     surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, w, h);
     cairo = cairo_create (surface);
 
-    cairo_scale(cairo, scale, scale);
+    cairo_scale(cairo, self->priv->scale_x, self->priv->scale_y);
     cairo_set_source_surface (cairo, priv->cursor->data, 0, 0);
     cairo_paint (cairo);
 
@@ -153,133 +155,6 @@ frdp_session_update_mouse_pointer (FrdpSession  *self)
   gdk_window_set_cursor (window, cursor);
 }
 
-static BOOL
-frdp_Pointer_New(rdpContext* context, rdpPointer* pointer)
-{
-  frdpContext *fcontext = (frdpContext*) context;
-  frdpPointer *fpointer = (frdpPointer*) pointer;
-  int stride;
-	unsigned char *data;
-  cairo_surface_t *surface;
-
-	if (!fcontext || !fpointer)
-		return FALSE;
-
-  surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, pointer->width,
-                                        pointer->height);
-  if (!surface) {
-    return FALSE;
-  }
-
-  { /* FreeRDP BUG https://github.com/FreeRDP/FreeRDP/issues/5061
-     * the function freerdp_image_copy_from_pointer_data
-     * does not initialize the buffer which results in broken alpha data. */
-    cairo_t* cairo = cairo_create (surface);
-
-    cairo_set_source_rgba (cairo, 0.0, 0.0, 0.0, 1.0);
-    cairo_fill (cairo);
-    cairo_paint (cairo);
-    cairo_destroy (cairo);
-  }
-
-  data = cairo_image_surface_get_data (surface);
-  if (!data) {
-    goto fail;
-  }
-
-  stride = cairo_format_stride_for_width (CAIRO_FORMAT_ARGB32, pointer->width);
-	if (!freerdp_image_copy_from_pointer_data (data, PIXEL_FORMAT_BGRA32,
-                                             stride, 0, 0, pointer->width,
-                                             pointer->height,
-                                             pointer->xorMaskData,
-                                             pointer->lengthXorMask,
-                                             pointer->andMaskData,
-                                             pointer->lengthAndMask,
-                                             pointer->xorBpp,
-                                             &context->gdi->palette))
-    goto fail;
-
-  fpointer->data = surface;
-  return TRUE;
-fail:
-  if (surface)
-    cairo_surface_destroy (surface);
-	return FALSE;
-}
-
-static void
-frdp_Pointer_Free (rdpContext* context, rdpPointer* pointer)
-{
-  frdpPointer *fpointer = (frdpPointer*) pointer;
-
-  if (fpointer && fpointer->data) {
-    cairo_surface_destroy (fpointer->data);
-    fpointer->data = NULL;
-  }
-}
-
-static BOOL
-frdp_Pointer_Set (rdpContext* context,
-                  const rdpPointer* pointer)
-{
-  frdpContext *fcontext = (frdpContext*) context;
-  frdpPointer *fpointer = (frdpPointer*) pointer;
-  FrdpSessionPrivate *priv = fcontext->self->priv;
-
-  priv->cursor = fpointer;
-  priv->cursor_null = FALSE;
-
-  frdp_session_update_mouse_pointer (fcontext->self);
-	return TRUE;
-}
-
-static BOOL
-frdp_Pointer_SetNull (rdpContext* context)
-{
-  frdpContext *fcontext = (frdpContext*) context;
-  FrdpSessionPrivate *priv = fcontext->self->priv;
-
-  priv->cursor = NULL;
-  priv->cursor_null = TRUE;
-
-  frdp_session_update_mouse_pointer (fcontext->self);
-
-  return TRUE;
-}
-
-static BOOL
-frdp_Pointer_SetDefault (rdpContext* context)
-{
-  frdpContext *fcontext = (frdpContext*) context;
-  FrdpSessionPrivate *priv = fcontext->self->priv;
-
-  priv->cursor = NULL;
-  priv->cursor_null = FALSE;
-  frdp_session_update_mouse_pointer (fcontext->self);
-	return TRUE;
-}
-
-static BOOL
-frdp_Pointer_SetPosition (rdpContext* context, UINT32 x, UINT32 y)
-{
-  return TRUE;
-}
-
-static void
-frdp_register_pointer (rdpGraphics* graphics)
-{
-	rdpPointer pointer;
-
-	pointer.size = sizeof(frdpPointer);
-	pointer.New = frdp_Pointer_New;
-	pointer.Free = frdp_Pointer_Free;
-	pointer.Set = frdp_Pointer_Set;
-	pointer.SetNull = frdp_Pointer_SetNull;
-	pointer.SetDefault = frdp_Pointer_SetDefault;
-	pointer.SetPosition = frdp_Pointer_SetPosition;
-	graphics_register_pointer(graphics, &pointer);
-}
-
 static guint32
 frdp_session_get_best_color_depth (FrdpSession *self)
 {
@@ -292,31 +167,69 @@ frdp_session_get_best_color_depth (FrdpSession *self)
   return gdk_visual_get_depth (visual);
 }
 
+static void
+create_cairo_surface (FrdpSession *self)
+{
+  FrdpSessionPrivate *priv = self->priv;
+  rdpSettings *settings;
+  rdpGdi *gdi;
+  gint stride;
+
+  if (priv->surface != NULL) {
+    cairo_surface_mark_dirty (priv->surface);
+    cairo_surface_destroy (priv->surface);
+    self->priv->surface = NULL;
+  }
+
+  gdi = priv->freerdp_session->context->gdi;
+  settings = priv->freerdp_session->context->settings;
+  gdi_resize (gdi, settings->DesktopWidth, settings->DesktopHeight);
+
+  gtk_widget_set_size_request (priv->display,
+                               settings->DesktopWidth,
+                               settings->DesktopHeight);
+  stride = cairo_format_stride_for_width (priv->cairo_format, gdi->width);
+  self->priv->surface =
+      cairo_image_surface_create_for_data ((unsigned char*) gdi->primary_buffer,
+                                           priv->cairo_format,
+                                           gdi->width,
+                                           gdi->height,
+                                           stride);
+  cairo_surface_flush (priv->surface);
+}
+
 static void
 frdp_session_configure_event (GtkWidget *widget,
                               GdkEvent  *event,
                               gpointer   user_data)
 {
   FrdpSession *self = (FrdpSession*) user_data;
-  rdpSettings *settings = self->priv->freerdp_session->settings;
+  FrdpSessionPrivate *priv = self->priv;
+  GtkScrolledWindow *scrolled;
+  rdpSettings *settings = priv->freerdp_session->settings;
+  rdpGdi *gdi;
   double width, height;
 
-  if (self->priv->scaling) {
-    width = (double)gtk_widget_get_allocated_width (widget);
-    height = (double)gtk_widget_get_allocated_height (widget);
-
-    if (width < height)
-      self->priv->scale = width / settings->DesktopWidth;
-    else
-      self->priv->scale = height / settings->DesktopHeight;
-
-    settings->DesktopScaleFactor = self->priv->scale;
-
-    self->priv->offset_x = (width - settings->DesktopWidth * self->priv->scale) / 2.0;
-    self->priv->offset_y = (height - settings->DesktopHeight * self->priv->scale) / 2.0;
+  gdi = priv->freerdp_session->context->gdi;
+  if (priv->surface == NULL) {
+    create_cairo_surface (self);
   }
 
-  frdp_session_update_mouse_pointer (self);
+  scrolled = gtk_widget_get_ancestor (widget, GTK_TYPE_SCROLLED_WINDOW);
+  width = (double)gtk_widget_get_allocated_width (scrolled);
+  height = (double)gtk_widget_get_allocated_height (scrolled);
+
+  if (priv->scaling) {
+    priv->scale_x = width / settings->DesktopWidth;
+    priv->scale_y = height / settings->DesktopHeight;
+
+    priv->offset_x = (width - settings->DesktopWidth * priv->scale_x) / 2.0;
+    priv->offset_y = (height - settings->DesktopHeight * priv->scale_y) / 2.0;
+
+    gtk_widget_set_size_request (priv->display, settings->DesktopWidth, settings->DesktopHeight);
+  } else {
+    gtk_widget_set_size_request (priv->display, gdi->width, gdi->height);
+  }
 }
 
 static void
@@ -324,8 +237,6 @@ frdp_session_set_scaling (FrdpSession *self,
                           gboolean     scaling)
 {
   self->priv->scaling = scaling;
-
-  frdp_session_configure_event (self->priv->display, NULL, self);
 }
 
 static gboolean
@@ -335,13 +246,18 @@ frdp_session_draw (GtkWidget *widget,
 {
   FrdpSession *self = (FrdpSession*) user_data;
 
+  create_cairo_surface (self);
+
   if (self->priv->scaling) {
       cairo_translate (cr, self->priv->offset_x, self->priv->offset_y);
-      cairo_scale (cr, self->priv->scale, self->priv->scale);
+      cairo_scale (cr, self->priv->scale_x, self->priv->scale_y);
   }
+
   cairo_set_source_surface (cr, self->priv->surface, 0, 0);
   cairo_paint (cr);
 
+  frdp_display_set_scaling (self->priv->display, self->priv->scaling);
+
   return TRUE;
 }
 
@@ -422,13 +338,13 @@ frdp_end_paint (rdpContext *context)
   priv = self->priv;
 
   if (priv->scaling) {
-      pos_x = self->priv->offset_x + x * priv->scale;
-      pos_y = self->priv->offset_y + y * priv->scale;
+      pos_x = self->priv->offset_x + x * priv->scale_x;
+      pos_y = self->priv->offset_y + y * priv->scale_y;
       gtk_widget_queue_draw_area (priv->display,
                                   floor (pos_x),
                                   floor (pos_y),
-                                  ceil (pos_x + w * priv->scale) - floor (pos_x),
-                                  ceil (pos_y + h * priv->scale) - floor (pos_y));
+                                  ceil (pos_x + w * priv->scale_x) - floor (pos_x),
+                                  ceil (pos_y + h * priv->scale_y) - floor (pos_y));
   } else {
     gtk_widget_queue_draw_area (priv->display, x, y, w, h);
   }
@@ -440,53 +356,33 @@ static gboolean
 frdp_post_connect (freerdp *freerdp_session)
 {
   FrdpSession *self = ((frdpContext *) freerdp_session->context)->self;
-  cairo_format_t cairo_format;
-  rdpGdi *gdi;
   guint32 color_format;
-  gint stride;
 
   switch (frdp_session_get_best_color_depth (self)) {
     case 32:
       color_format = PIXEL_FORMAT_BGRA32;
-      cairo_format = CAIRO_FORMAT_ARGB32;
+      self->priv->cairo_format = CAIRO_FORMAT_ARGB32;
       break;
     case 24:
       color_format = PIXEL_FORMAT_BGRX32;
-      cairo_format = CAIRO_FORMAT_RGB24;
+      self->priv->cairo_format = CAIRO_FORMAT_RGB24;
       break;
     case 16:
     case 15:
       color_format = PIXEL_FORMAT_BGR16;
-      cairo_format = CAIRO_FORMAT_RGB16_565;
+      self->priv->cairo_format = CAIRO_FORMAT_RGB16_565;
       break;
     default:
       color_format = PIXEL_FORMAT_BGRX32;
-      cairo_format = CAIRO_FORMAT_RGB16_565;
+      self->priv->cairo_format = CAIRO_FORMAT_RGB16_565;
       break;
   }
 
   gdi_init (freerdp_session, color_format);
-  gdi = freerdp_session->context->gdi;
 
-  frdp_register_pointer (freerdp_session->context->graphics);
-  pointer_cache_register_callbacks(freerdp_session->context->update);
   freerdp_session->update->BeginPaint = frdp_begin_paint;
   freerdp_session->update->EndPaint = frdp_end_paint;
 
-  stride = cairo_format_stride_for_width (cairo_format, gdi->width);
-  self->priv->surface =
-      cairo_image_surface_create_for_data ((unsigned char*) gdi->primary_buffer,
-                                           cairo_format,
-                                           gdi->width,
-                                           gdi->height,
-                                           stride);
-
-  gtk_widget_queue_draw_area (self->priv->display,
-                              0,
-                              0,
-                              gdi->width,
-                              gdi->height);
-
   return TRUE;
 }
 
@@ -504,7 +400,6 @@ idle_close (gpointer user_data)
 
   if (self->priv->freerdp_session != NULL) {
     freerdp_disconnect (self->priv->freerdp_session);
-    freerdp_context_free (self->priv->freerdp_session);
     g_clear_pointer (&self->priv->freerdp_session, freerdp_free);
   }
 
@@ -543,8 +438,11 @@ update (gpointer user_data)
     return FALSE;
 
   if (!freerdp_check_event_handles (priv->freerdp_session->context)) {
+    if (freerdp_get_last_error(priv->freerdp_session->context) == FREERDP_ERROR_SUCCESS) {
       g_warning ("Failed to check FreeRDP file descriptor");
-      return FALSE;
+    }
+
+    return TRUE;
   }
 
   if (freerdp_shall_disconnect (priv->freerdp_session)) {
@@ -594,6 +492,10 @@ frdp_session_init_freerdp (FrdpSession *self)
   settings->UseRdpSecurityLayer = FALSE;
 
   settings->NegotiateSecurityLayer = TRUE;
+
+  settings->DesktopResize = TRUE;
+  settings->DynamicResolutionUpdate = TRUE;
+  settings->SupportDisplayControl = TRUE;
 }
 
 static void
@@ -632,7 +534,11 @@ frdp_session_connect_thread (GTask        *task,
             break;
 
         default:
-            g_warning ("Unhandled FreeRDP error: '%s'",
+            g_signal_emit (self,
+                           signals[RDP_ERROR], 0,
+                           freerdp_get_last_error_string (error_code));
+
+            g_warning ("Unexpected RDP error: '%s'",
                        freerdp_get_last_error_string (error_code));
             break;
     }
@@ -643,11 +549,12 @@ frdp_session_connect_thread (GTask        *task,
     return;
   }
 
+  gtk_widget_realize (self->priv->display);
+  create_cairo_surface (self);
   g_signal_connect (self->priv->display, "draw",
                     G_CALLBACK (frdp_session_draw), self);
   g_signal_connect (self->priv->display, "configure-event",
                     G_CALLBACK (frdp_session_configure_event), self);
-  frdp_session_set_scaling (self, TRUE);
 
   self->priv->update_id = g_idle_add ((GSourceFunc) update, self);
 
@@ -793,23 +700,31 @@ frdp_session_class_init (FrdpSessionClass *klass)
                                                          TRUE,
                                                          G_PARAM_READWRITE));
 
+  signals[RDP_ERROR] = g_signal_new ("rdp-error",
+                                     FRDP_TYPE_SESSION,
+                                     G_SIGNAL_RUN_FIRST,
+                                     0, NULL, NULL, NULL,
+                                     G_TYPE_NONE, 1,
+                                     G_TYPE_STRING);
+
   signals[RDP_CONNECTED] = g_signal_new ("rdp-connected",
                                          FRDP_TYPE_SESSION,
                                          G_SIGNAL_RUN_FIRST,
                                          0, NULL, NULL, NULL,
                                          G_TYPE_NONE, 0);
+
   signals[RDP_DISCONNECTED] = g_signal_new ("rdp-disconnected",
                                             FRDP_TYPE_SESSION,
                                             G_SIGNAL_RUN_FIRST,
                                             0, NULL, NULL, NULL,
                                             G_TYPE_NONE, 0);
+
   signals[RDP_AUTH_FAILURE] = g_signal_new ("rdp-auth-failure",
                                             FRDP_TYPE_SESSION,
                                             G_SIGNAL_RUN_FIRST,
                                             0, NULL, NULL, NULL,
                                             G_TYPE_NONE, 1,
                                             G_TYPE_STRING);
-
 }
 
 static void
@@ -919,8 +834,8 @@ frdp_session_mouse_event (FrdpSession          *self,
   input = priv->freerdp_session->input;
 
   if (priv->scaling) {
-    x = (x - priv->offset_x) / priv->scale;
-    y = (y - priv->offset_y) / priv->scale;
+    x = (x - priv->offset_x) / priv->scale_x;
+    y = (y - priv->offset_y) / priv->scale_y;
   }
 
   x = x < 0.0 ? 0.0 : x;
